from __future__ import annotations

from typing import Sequence

import einops
import torch

from torch_fourier_shift.dft_utils import rfft_shape

def fftfreq_grid_1d(
    image_shape: tuple[int],
    rfft: bool,
    spacing: float = 1,
    device: torch.device = None
) -> torch.Tensor:
    """Construct a grid of DFT sample freqs for a 1D image.

    Parameters
    ----------
    image_shape: int
        A 1D shape `(w, )` of the input image for which a grid of DFT sample freqs
        should be calculated.
    rfft: bool
        Whether the frequency grid is for a real fft (rfft).
    spacing: float
        Sample spacing in `w` dimension of the grid.
    device: torch.device
        Torch device for the resulting grid.

    Returns
    -------
    frequency_grid: torch.Tensor
        `(w, )` array of DFT sample frequencies.
    """
    dw = spacing
    frequency_func = torch.fft.rfftfreq if rfft is True else torch.fft.fftfreq
    (w, ) = image_shape
    freq_x = frequency_func(w, d=dw, device=device)
    return freq_x

def fftfreq_grid_2d(
    image_shape: tuple[int, int],
    rfft: bool,
    spacing: float | tuple[float, float] = 1,
    device: torch.device = None
) -> torch.Tensor:
    """Construct a grid of DFT sample freqs for a 2D image.

    Parameters
    ----------
    image_shape: Sequence[int]
        A 2D shape `(h, w)` of the input image for which a grid of DFT sample freqs
        should be calculated.
    rfft: bool
        Whether the frequency grid is for a real fft (rfft).
    spacing: float | Tuple[float, float]
        Sample spacing in `h` and `w` dimensions of the grid.
    device: torch.device
        Torch device for the resulting grid.

    Returns
    -------
    frequency_grid: torch.Tensor
        `(h, w, 2)` array of DFT sample freqs.
        Order of freqs in the last dimension corresponds to the order of
        the two dimensions of the grid.
    """
    dh, dw = spacing if isinstance(spacing, Sequence) else [spacing] * 2
    last_axis_frequency_func = torch.fft.rfftfreq if rfft is True else torch.fft.fftfreq
    h, w = image_shape
    freq_y = torch.fft.fftfreq(h, d=dh, device=device)
    freq_x = last_axis_frequency_func(w, d=dw, device=device)
    h, w = rfft_shape(image_shape) if rfft is True else image_shape
    freq_yy = einops.repeat(freq_y, 'h -> h w', w=w)
    freq_xx = einops.repeat(freq_x, 'w -> h w', h=h)
    return einops.rearrange([freq_yy, freq_xx], 'freq h w -> h w freq')


def fftfreq_grid_3d(
    image_shape: Sequence[int],
    rfft: bool,
    spacing: float | tuple[float, float, float] = 1,
    device: torch.device = None
) -> torch.Tensor:
    """Construct a grid of DFT sample freqs for a 3D image.

    Parameters
    ----------
    image_shape: Sequence[int]
        A 3D shape `(d, h, w)` of the input image for which a grid of DFT sample freqs
        should be calculated.
    rfft: bool
        Controls Whether the frequency grid is for a real fft (rfft).
    spacing: float | Tuple[float, float, float]
        Sample spacing in `d`, `h` and `w` dimensions of the grid.
    device: torch.device
        Torch device for the resulting grid.

    Returns
    -------
    frequency_grid: torch.Tensor
        `(h, w, 3)` array of DFT sample freqs.
        Order of freqs in the last dimension corresponds to the order of dimensions
        of the grid.
    """
    dd, dh, dw = spacing if isinstance(spacing, Sequence) else [spacing] * 3
    last_axis_frequency_func = torch.fft.rfftfreq if rfft is True else torch.fft.fftfreq
    d, h, w = image_shape
    freq_z = torch.fft.fftfreq(d, d=dd, device=device)
    freq_y = torch.fft.fftfreq(h, d=dh, device=device)
    freq_x = last_axis_frequency_func(w, d=dw, device=device)
    d, h, w = rfft_shape(image_shape) if rfft is True else image_shape
    freq_zz = einops.repeat(freq_z, 'd -> d h w', h=h, w=w)
    freq_yy = einops.repeat(freq_y, 'h -> d h w', d=d, w=w)
    freq_xx = einops.repeat(freq_x, 'w -> d h w', d=d, h=h)
    return einops.rearrange([freq_zz, freq_yy, freq_xx], 'freq ... -> ... freq')



from typing import Sequence

import torch


def rfft_shape(input_shape: Sequence[int]) -> tuple[int, ...]:
    """Get the output shape of an rfft on an input with input_shape."""
    rfft_shape = list(input_shape)
    rfft_shape[-1] = int((rfft_shape[-1] / 2) + 1)
    return tuple(rfft_shape)

def fftshift_1d(input: torch.Tensor, rfft: bool = True) -> torch.Tensor:
    if rfft is False:
        output = torch.fft.fftshift(input, dim=(-1, ))
    else:
        output = input
    return output

def fftshift_2d(input: torch.Tensor, rfft: bool) -> torch.Tensor:
    if rfft is False:
        output = torch.fft.fftshift(input, dim=(-2, -1))
    else:
        output = torch.fft.fftshift(input, dim=(-2,))
    return output

def ifftshift_1d(input: torch.Tensor, rfft: bool) -> torch.Tensor:
    if rfft is False:
        output = torch.fft.ifftshift(input, dim=(-1, ))
    else:
        output = input
    return output

def ifftshift_2d(input: torch.Tensor, rfft: bool) -> torch.Tensor:
    if rfft is False:
        output = torch.fft.ifftshift(input, dim=(-2, -1))
    else:
        output = torch.fft.ifftshift(input, dim=(-2,))
    return output


def fftshift_3d(input: torch.Tensor, rfft: bool) -> torch.Tensor:
    if rfft is False:
        output = torch.fft.fftshift(input, dim=(-3, -2, -1))
    else:
        output = torch.fft.fftshift(input, dim=(-3, -2,))
    return output


def ifftshift_3d(input: torch.Tensor, rfft: bool) -> torch.Tensor:
    if rfft is False:
        output = torch.fft.ifftshift(input, dim=(-3, -2, -1))
    else:
        output = torch.fft.ifftshift(input, dim=(-3, -2,))
    return output



import torch

from torch_fourier_shift import fourier_shift_image_3d
from torch_fourier_shift.phase_shift_grids import phase_shift_grid_3d


def test_get_phase_shifts_3d_full_fft():
    shifts = torch.zeros(size=(1, 3))
    phase_shifts = phase_shift_grid_3d(shifts, image_shape=(2, 2, 2), rfft=False)
    assert torch.allclose(
        phase_shifts, torch.ones(size=(2, 2, 2), dtype=torch.complex64)
    )

    shifts = torch.tensor([[1, 2, 3]])
    phase_shifts = phase_shift_grid_3d(shifts, image_shape=(2, 2, 2), rfft=False)
    expected = torch.tensor([[[1. + 0.0000e+00j, -1. - 2.3850e-08j],
                              [1. + 1.7485e-07j, -1. - 6.7553e-07j]],
                             [[-1. - 8.7423e-08j, 1. + 3.4969e-07j],
                              [-1. - 2.3850e-08j, 1. + 4.7700e-08j]]])
    assert torch.allclose(phase_shifts, expected)


def test_get_phase_shifts_3d_rfft():
    shifts = torch.zeros(size=(1, 3))
    phase_shifts = phase_shift_grid_3d(shifts, image_shape=(2, 2, 2), rfft=True)
    assert phase_shifts.shape == (1, 2, 2, 2)
    expected = torch.ones(size=(2, 2, 2), dtype=torch.complex64)
    assert torch.allclose(phase_shifts, expected)

    shifts = torch.tensor([[1, 2, 3]])
    phase_shifts = phase_shift_grid_3d(shifts, image_shape=(2, 2, 2), rfft=True)
    expected = torch.tensor([[[1. + 0.0000e+00j, -1. + 2.3850e-08j],
                              [1. + 1.7485e-07j, -1. - 1.5100e-07j]],
                             [[-1. - 8.7423e-08j, 1. + 3.0199e-07j],
                              [-1. - 2.3850e-08j, 1. + 0.0000e+00j]]])
    assert torch.allclose(phase_shifts, expected)


def test_fourier_shift_image_3d():
    image = torch.zeros((4, 4, 4))
    image[2, 2, 2] = 1

    # +1px in each dimension
    shifts = torch.ones((1, 3))
    shifted = fourier_shift_image_3d(image, shifts)
    expected = torch.zeros((4, 4, 4))
    expected[3, 3, 3] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)

    # +1px in z
    shifts = torch.zeros((1, 3))
    shifts[0, 0] = 1
    shifted = fourier_shift_image_3d(image, shifts)
    expected = torch.zeros((4, 4, 4))
    expected[3, 2, 2] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)

    # +1px in y
    shifts = torch.zeros((1, 3))
    shifts[0, 1] = 1
    shifted = fourier_shift_image_3d(image, shifts)
    expected = torch.zeros((4, 4, 4))
    expected[2, 3, 2] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)

    # +1px in x
    shifts = torch.zeros((1, 3))
    shifts[0, 2] = 1
    shifted = fourier_shift_image_3d(image, shifts)
    expected = torch.zeros((4, 4, 4))
    expected[2, 2, 3] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)


def test_fourier_shift_preserves_dimensions_3d():
    """Test that fourier_shift_image_1d preserves input dimensions."""
    image_3d_odd = torch.randn((127, 127, 127))
    shifts = torch.tensor((5.0, 5.0, 5.0))
    shifted_3d_odd = fourier_shift_image_3d(image_3d_odd, shifts)
    assert shifted_3d_odd.shape == image_3d_odd.shape, (
        f"3D odd: Expected {image_3d_odd.shape}, got {shifted_3d_odd.shape}"
    )



import torch

from .phase_shift_grids import (
    phase_shift_grid_1d,
    phase_shift_grid_2d,
    phase_shift_grid_3d,
)


def fourier_shift_dft_1d(
    dft: torch.Tensor,
    image_shape: tuple[int],
    shifts: torch.Tensor,
    rfft: bool,
    fftshifted: bool,
    cache_intermediates: bool = False,
):
    """Apply phase shifts to 2D discrete Fourier transforms.

    Parameters
    ----------
    dft: torch.Tensor
        `(..., w)` array containing DFTs.
    image_shape: tuple[int]
        `(w, )` of images prior to DFT computation.
    shifts: torch.Tensor
        `(...)` array of 2D shifts in `w`.
    rfft: bool
        Whether the input was computed using `rfft`.
    fftshifted: bool
        Whether the DFTs have been fftshifted to center the DC component.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached. By default, we use an LRU cache with size=3. You can change the size of the cache by setting the environmental variable TORCH_FOURIER_SHIFT_CACHE_SIZE before importing the module.

    Returns
    -------
    shifted_dfts: torch.Tensor
        `(..., w)` array of DFTs with phase shifts applied.
    """
    phase_shifts = phase_shift_grid_1d(
        shifts=shifts,
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshifted,
        cache_intermediates=cache_intermediates,
    )
    return dft * phase_shifts


def fourier_shift_dft_2d(
    dft: torch.Tensor,
    image_shape: tuple[int, int],
    shifts: torch.Tensor,
    rfft: bool,
    fftshifted: bool,
    cache_intermediates: bool = False,
):
    """Apply phase shifts to 2D discrete Fourier transforms.

    Parameters
    ----------
    dft: torch.Tensor
        `(..., h, w)` array containing DFTs.
    image_shape: tuple[int, int]
        `(h, w)` of images prior to DFT computation.
    shifts: torch.Tensor
        `(..., 2)` array of 2D shifts in `h` and `w`.
    rfft: bool
        Whether the input was computed using `rfft`.
    fftshifted: bool
        Whether the DFTs have been fftshifted to center the DC component.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached. By default, we use an LRU cache with size=3. You can change the size of the cache by setting the environmental variable TORCH_FOURIER_SHIFT_CACHE_SIZE before importing the module.

    Returns
    -------
    shifted_dfts: torch.Tensor
        `(..., h, w)` array of DFTs with phase shifts applied.
    """
    phase_shifts = phase_shift_grid_2d(
        shifts=shifts,
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshifted,
        cache_intermediates=cache_intermediates,
    )
    return dft * phase_shifts


def fourier_shift_dft_3d(
    dft: torch.Tensor,
    image_shape: tuple[int, int, int],
    shifts: torch.Tensor,
    rfft: bool = False,
    fftshifted: bool = False,
    cache_intermediates: bool = False,
):
    """Apply phase shifts to 3D discrete Fourier transforms.

    Parameters
    ----------
    dft: torch.Tensor
        `(..., h, w)` array containing DFTs.
    image_shape: tuple[int, int, int]
        `(h, w)` of images prior to DFT computation.
    shifts: torch.Tensor
        `(..., 3)` array of 3D shifts in `d`, `h` and `w`.
    rfft: bool
        Whether the input was computed using `rfft`.
    fftshifted: bool
        Whether the DFTs have been fftshifted to center the DC component.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached. By default, we use an LRU cache with size=3. You can change the size of the cache by setting the environmental variable TORCH_FOURIER_SHIFT_CACHE_SIZE before importing the module.
        
    Returns
    -------
    shifted_dfts: torch.Tensor
        `(..., h, w)` array of DFTs with phase shifts applied.
    """
    phase_shifts = phase_shift_grid_3d(
        shifts=shifts,
        image_shape=image_shape,
        rfft=rfft,
        fftshift=fftshifted,
        cache_intermediates=cache_intermediates,
    )
    return dft * phase_shifts



import torch

from torch_fourier_shift.fourier_shift_image import fourier_shift_image_1d
from torch_fourier_shift.phase_shift_grids import phase_shift_grid_1d


def test_get_phase_shifts_1d_full_fft():
    shifts = torch.zeros(size=(1, 2))
    phase_shifts = phase_shift_grid_1d(shifts, image_shape=(2,), rfft=False)
    assert torch.allclose(phase_shifts, torch.ones(size=(2,), dtype=torch.complex64))

    shifts = torch.tensor([1, 2])
    phase_shifts = phase_shift_grid_1d(shifts, image_shape=(2,), rfft=False)
    expected = torch.tensor([[1. - 0.0000e+00j, -1. - 8.7423e-08j],
                             [1. - 0.0000e+00j, 1. + 1.7485e-07j]])
    assert torch.allclose(phase_shifts, expected)


def test_get_phase_shifts_1d_rfft():
    shifts = torch.zeros(size=(1, 2))
    phase_shifts = phase_shift_grid_1d(shifts, image_shape=(2,), rfft=True)
    assert phase_shifts.shape == (1, 2, 2)
    expected = torch.ones(size=(1, 2, 2), dtype=torch.complex64)
    assert torch.allclose(phase_shifts, expected)

    shifts = torch.tensor([[1, 2]])
    phase_shifts = phase_shift_grid_1d(shifts, image_shape=(2,), rfft=False)
    expected = torch.tensor([[[1. - 0.0000e+00j, -1. - 8.7423e-08j],
                              [1. - 0.0000e+00j, 1. + 1.7485e-07j]]])
    assert torch.allclose(phase_shifts, expected)


def test_fourier_shift_image_1d():
    image = torch.zeros((4,))
    image[2] = 1

    # +1px
    shifts = torch.ones((1,))
    shifted = fourier_shift_image_1d(image, shifts)
    expected = torch.zeros((4,))
    expected[3] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)

    # -1px
    shifts = -1 * torch.ones((1,))
    shifted = fourier_shift_image_1d(image, shifts)
    expected = torch.zeros((4,))
    expected[1] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)
    



def test_fourier_shift_preserves_dimensions_1d():
    """Test that fourier_shift_image_1d preserves input dimensions."""
    image_1d_odd = torch.randn(127)
    shifts = torch.tensor(5.0)
    shifted_1d_odd = fourier_shift_image_1d(image_1d_odd, shifts)
    assert shifted_1d_odd.shape == image_1d_odd.shape, (
        f"1D odd: Expected {image_1d_odd.shape}, got {shifted_1d_odd.shape}"
    )




import torch

from torch_fourier_shift.phase_shift_grids import (
    phase_shift_grid_2d,
    fftfreq_grid_2d_with_cache,
)


def test_cache_2d():
    """Test caching for 2D phase shift grids."""
    shape = (16, 16)
    shifts = torch.randn(1, 2)
    
    # Clear cache before test
    fftfreq_grid_2d_with_cache.cache_clear()

    # Call once to populate the cache
    phase_shift_grid_2d(
        shifts, image_shape=shape, rfft=False, fftshift=False, cache_intermediates=True
    )
    assert fftfreq_grid_2d_with_cache.cache_info().hits == 0
    assert fftfreq_grid_2d_with_cache.cache_info().misses == 1
    
    # Call again to get a cache hit
    phase_shift_grid_2d(
        shifts, image_shape=shape, rfft=False, fftshift=False, cache_intermediates=True
    )
    assert fftfreq_grid_2d_with_cache.cache_info().hits == 1
    assert fftfreq_grid_2d_with_cache.cache_info().misses == 1


def test_cached_vs_uncached_2d():
    """Test that cached and uncached 2D phase shift grids are the same."""
    shape = (16, 16)
    shifts = torch.randn(1, 2)

    # Get cached result
    cached_result = phase_shift_grid_2d(
        shifts, image_shape=shape, rfft=False, fftshift=False, cache_intermediates=True
    )

    # Get uncached result
    uncached_result = phase_shift_grid_2d(
        shifts, image_shape=shape, rfft=False, fftshift=False, cache_intermediates=False
    )

    torch.testing.assert_close(cached_result, uncached_result)



from .fourier_shift_dft import fourier_shift_dft_1d, fourier_shift_dft_2d, fourier_shift_dft_3d
from .fourier_shift_image import fourier_shift_image_1d, fourier_shift_image_2d, fourier_shift_image_3d



import torch

from torch_fourier_shift import fourier_shift_image_2d
from torch_fourier_shift.phase_shift_grids import phase_shift_grid_2d


def test_get_phase_shifts_2d_full_fft():
    shifts = torch.zeros(size=(1, 2))
    phase_shifts = phase_shift_grid_2d(shifts, image_shape=(2, 2), rfft=False)
    assert torch.allclose(phase_shifts, torch.ones(size=(2, 2), dtype=torch.complex64))

    shifts = torch.tensor([[1, 2]])
    phase_shifts = phase_shift_grid_2d(shifts, image_shape=(2, 2), rfft=False)
    expected = torch.tensor([[[1 + 0.0000e+00j, 1 + 1.7485e-07j],
                              [-1 - 8.7423e-08j, -1 - 2.3850e-08j]]])
    assert torch.allclose(phase_shifts, expected)


def test_get_phase_shifts_2d_rfft():
    shifts = torch.zeros(size=(1, 2))
    phase_shifts = phase_shift_grid_2d(shifts, image_shape=(2, 2), rfft=True)
    assert phase_shifts.shape == (1, 2, 2)
    expected = torch.ones(size=(2, 2), dtype=torch.complex64)
    assert torch.allclose(phase_shifts, expected)

    shifts = torch.tensor([[1, 2]])
    phase_shifts = phase_shift_grid_2d(shifts, image_shape=(2, 2), rfft=False)
    expected = torch.tensor([[[1 + 0.0000e+00j, 1 + 1.7485e-07j],
                              [-1 - 8.7423e-08j, -1 - 2.3850e-08j]]])
    assert torch.allclose(phase_shifts, expected)


def test_fourier_shift_image_2d():
    image = torch.zeros((4, 4))
    image[2, 2] = 1

    # +1px in each dimension
    shifts = torch.ones((1, 2))
    shifted = fourier_shift_image_2d(image, shifts)
    expected = torch.zeros((4, 4))
    expected[3, 3] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)

    # +1px in y
    shifts = torch.zeros((1, 2))
    shifts[0, 0] = 1
    shifted = fourier_shift_image_2d(image, shifts)
    expected = torch.zeros((4, 4))
    expected[3, 2] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)

    # +1px in x
    shifts = torch.zeros((1, 2))
    shifts[0, 1] = 1
    shifted = fourier_shift_image_2d(image, shifts)
    expected = torch.zeros((4, 4))
    expected[2, 3] = 1
    assert torch.allclose(shifted, expected, atol=1e-5)


def test_fourier_shift_preserves_dimensions_2d():
    """Test that fourier_shift_image_1d preserves input dimensions."""
    image_2d_odd = torch.randn((127, 127))
    shifts = torch.tensor((5.0, 5.0))
    shifted_2d_odd = fourier_shift_image_2d(image_2d_odd, shifts)
    assert shifted_2d_odd.shape == image_2d_odd.shape, (
        f"2D odd: Expected {image_2d_odd.shape}, got {shifted_2d_odd.shape}"
    )



import torch

from torch_fourier_shift.fourier_shift_dft import (
    fourier_shift_dft_1d,
    fourier_shift_dft_2d,
    fourier_shift_dft_3d,
)


def fourier_shift_image_1d(image: torch.Tensor, shifts: torch.Tensor,
    cache_intermediates: bool = False
):
    """Translate one or more 1D images by phase shifting their Fourier transforms.

    Parameters
    ----------
    image: torch.Tensor
        `(..., w)` image(s).
    shifts: torch.Tensor
        `(..., )` array of 1D shifts in `w`.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached
        
    Returns
    -------
    shifted_images: torch.Tensor
        `(..., w)` array of shifted images.
    """
    w = image.shape[-1]

    image = torch.fft.rfftn(image, dim=(-1,))
    image = fourier_shift_dft_1d(
        image,
        image_shape=(w,),
        shifts=shifts,
        rfft=True,
        fftshifted=False,
        cache_intermediates=cache_intermediates,
    )
    image = torch.fft.irfftn(image, dim=(-1,), s=(w,))
    return torch.real(image)


def fourier_shift_image_2d(image: torch.Tensor, shifts: torch.Tensor,
    cache_intermediates: bool = False
):
    """Translate one or more 2D images by phase shifting their Fourier transforms.

    Parameters
    ----------
    image: torch.Tensor
        `(..., h, w)` image(s).
    shifts: torch.Tensor
        `(..., 2)` array of 2D shifts in `h` and `w`.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached
        
    Returns
    -------
    shifted_images: torch.Tensor
        `(..., h, w)` array of shifted images.
    """
    h, w = image.shape[-2:]
    image = torch.fft.rfftn(image, dim=(-2, -1))
    image = fourier_shift_dft_2d(
        image,
        image_shape=(h, w),
        shifts=shifts,
        rfft=True,
        fftshifted=False,
        cache_intermediates=cache_intermediates,
    )
    image = torch.fft.irfftn(image, dim=(-2, -1), s=(h, w))
    return torch.real(image)


def fourier_shift_image_3d(image: torch.Tensor, shifts: torch.Tensor,
    cache_intermediates: bool = False
):
    """Translate one or more 3D images by phase shifting their Fourier transforms.

    Parameters
    ----------
    image: torch.Tensor
        `(..., d, h, w)` image(s).
    shifts: torch.Tensor
        `(..., 3)` array of 3D shifts in `d`, `h` and `w`.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached

    Returns
    -------
    shifted_image: torch.Tensor
        `(..., d, h, w)` array of shifted images.
    """
    d, h, w = image.shape[-3:]
    image = torch.fft.rfftn(image, dim=(-3, -2, -1))
    image = fourier_shift_dft_3d(
        image,
        image_shape=(d, h, w),
        shifts=shifts,
        rfft=True,
        fftshifted=False,
        cache_intermediates=cache_intermediates,
    )
    image = torch.fft.irfftn(image, dim=(-3, -2, -1), s=(d, h, w))
    return torch.real(image)



import os
import einops
import torch
from functools import lru_cache
from .fftfreq_grids import fftfreq_grid_2d, fftfreq_grid_3d, fftfreq_grid_1d
from .dft_utils import fftshift_1d, fftshift_2d, fftshift_3d

CACHE_SIZE = os.environ.get("TORCH_FOURIER_SHIFT_CACHE_SIZE", 3)

fftfreq_grid_1d_with_cache = lru_cache(maxsize=CACHE_SIZE)(fftfreq_grid_1d)
fftfreq_grid_2d_with_cache = lru_cache(maxsize=CACHE_SIZE)(fftfreq_grid_2d)
fftfreq_grid_3d_with_cache = lru_cache(maxsize=CACHE_SIZE)(fftfreq_grid_3d)

def phase_shift_grid_1d(
    shifts: torch.Tensor,
    image_shape: tuple[int],
    rfft: bool = False,
    fftshift: bool = False,
    cache_intermediates: bool = False
):
    """Generate arrays of values for phase shifting 1D DFTs by multiplication.

    Parameters
    ----------
    shifts: torch.Tensor
        `(..., )` array of shifts in the last image dimension.
    image_shape: tuple[int]
        Shape `(w, )` of 1D signal(s) on which phase shifts will be applied.
    rfft: bool
        If `True` the phase shifts generated will be compatible with
        the non-redundant half DFT outputs of the FFT for real inputs from `rfft`.
    fftshift: bool
        If `True`, fftshift the output.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached
    Returns
    -------
    phase_shifts: torch.Tensor
        `(..., w)` complex valued array of phase shifts for the fft or rfft
        of signals with `image_shape`.
    """
    
    if cache_intermediates:
        _fftfreq_grid_1d = fftfreq_grid_1d_with_cache
    else:
        _fftfreq_grid_1d = fftfreq_grid_1d
         
    fftfreq_grid = _fftfreq_grid_1d(
        image_shape=image_shape, rfft=rfft, device=shifts.device
    )  # (w, )
    shifts = einops.rearrange(shifts, '... -> ... 1')

    # radians/cycle * cycles/sample * samples = radians
    angles = -2 * torch.pi * fftfreq_grid * shifts  # (..., w)
    phase_shifts = torch.complex(real=torch.cos(angles), imag=torch.sin(angles))

    if fftshift is True:
        phase_shifts = fftshift_1d(phase_shifts, rfft=rfft)
    return phase_shifts


def phase_shift_grid_2d(
    shifts: torch.Tensor,
    image_shape: tuple[int, int],
    rfft: bool = False,
    fftshift: bool = False,
    cache_intermediates: bool = False
):
    """Generate arrays of values for phase shifting 2D DFTs by multiplication.

    Parameters
    ----------
    shifts: torch.Tensor
        `(..., 2)` array of shifts in the last two image dimensions `(h, w)`.
    image_shape: tuple[int, int]
        `(h, w)` of 2D image(s) on which phase shifts will be applied.
    rfft: bool
        If `True` the phase shifts generated will be compatible with
        the non-redundant half DFT outputs of the FFT for real inputs from `rfft`.
    fftshift: bool
        If `True`, fftshift the output.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached
    Returns
    -------
    phase_shifts: torch.Tensor
        `(..., h, w)` complex valued array of phase shifts for the fft or rfft
        of images with `image_shape`.
    """
    
    
    if cache_intermediates:
        _fftfreq_grid_2d = fftfreq_grid_2d_with_cache
    else:
        _fftfreq_grid_2d = fftfreq_grid_2d
        
    fftfreq_grid = _fftfreq_grid_2d(
        image_shape=image_shape, rfft=rfft, device=shifts.device
    )  # (h, w, 2)
    shifts = einops.rearrange(shifts, '... shift -> ... 1 1 shift')

    # radians/cycle * cycles/sample * samples = radians
    angles = einops.reduce(
        -2 * torch.pi * fftfreq_grid * shifts, '... h w 2 -> ... h w', reduction='sum'
    )
    phase_shifts = torch.complex(real=torch.cos(angles), imag=torch.sin(angles))
    if fftshift is True:
        phase_shifts = fftshift_2d(phase_shifts, rfft=rfft)
    return phase_shifts


def phase_shift_grid_3d(
    shifts: torch.Tensor,
    image_shape: tuple[int, int, int],
    rfft: bool = False,
    fftshift: bool = False,
    cache_intermediates: bool = False
):
    """Generate arrays of values for phase shifting 3D DFTs by multiplication.

    Parameters
    ----------
    shifts: torch.Tensor
        `(..., 3)` array of shifts in the last three image dimensions `(d, h, w)`.
    image_shape: tuple[int, int, int]
        `(d, h, w)` of 3D image(s) onto which phase shifts will be applied.
    rfft: bool
        If `True` the phase shifts generated will be compatible with
        the non-redundant half DFT outputs of the FFT for real inputs from `rfft`.
    fftshift: bool
        If `True`, fftshift the output.
    cache_intermediates: bool
        If `True`, the fftfreq_grid is cached

    Returns
    -------
    phase_shifts: torch.Tensor
        `(..., d, h, w)` complex valued array of phase shifts for the fft or rfft
        of images with `image_shape`.
    """
    
    
    if cache_intermediates:
        _fftfreq_grid_3d = fftfreq_grid_3d_with_cache
    else:
        _fftfreq_grid_3d = fftfreq_grid_3d
    
    fftfreq_grid = _fftfreq_grid_3d(
        image_shape=image_shape, rfft=rfft, device=shifts.device
    )  # (d, h, w, 3)
    shifts = einops.rearrange(shifts, '... shift -> ... 1 1 1 shift')

    # radians/cycle * cycles/sample * samples = radians
    angles = einops.reduce(
        -2 * torch.pi * fftfreq_grid * shifts, '... d h w 3 -> ... d h w', reduction='sum'
    )
    phase_shifts = torch.complex(real=torch.cos(angles), imag=torch.sin(angles))
    if fftshift is True:
        phase_shifts = fftshift_3d(phase_shifts, rfft=rfft)
    return phase_shifts
